"""
Parsers
-------

The Parser classes defined here are generally intended to be subclassed, with
user defined instance methods that correspond to a parsed item. The default
loader and reader simply return the arguments passed to their `load` and `read`
methods, so most use cases should use a different loader and reader.

>>> from soupstars import BaseParser
>>> parser = BaseParser(a=1, b=2)
>>> parser.load()  # returns the kwargs passed to BaseParser
{'a': 1, 'b': 2}
>>> parser.read()  # calls parser.load and returns the result
{'a': 1, 'b': 2}

Additionally, you can set a cache on the parser so that expensive load calls
are reduced. The default cache is just an in-memory dictionary, but caches
that use a local filesystem or Redis are available.

>>> parser.cache_id
'ac6dcde3d009f3b5d2ea803c12ef1a3c'
>>> parser.cache.get(parser.cache_id)
{'a': 1, 'b': 2}

Functions on the parsers decorated with `@parse` are considered "parsers" and
treated specially. They can be found by using the `iter_parsers` method.

>>> from soupstars import parse
>>> @parse
... def parse_a(read):
...     return read.get('a')
...
>>> parser.__class__.parse_a = parse_a
>>> parse_a in list(parser.iter_parsers())
True

Those parser functions should typically be accessed by treating the parser
instance like a dictionary.

>>> parser['parse_a']
1
"""

import collections
import doctest
import io
import json
import re

from flask import request, jsonify

from .readers import DefaultReader, HttpReader
from .loaders import DefaultLoader, HttpLoader
from .caches import DefaultCache
from .utils import SoupstarsPlugin


ParserTestResult = collections.namedtuple(
    'ParserTestResult',
    ['failed', 'message', 'parser']
)

ParserNameComponents = collections.namedtuple(
    'ParserNameComponents',
    ['first', 'last', 'classname']
)


class BaseParser(object):
    """
    The primary parsing object. Subclass this to create your own parser.
    """

    loader = DefaultLoader
    reader = DefaultReader
    cache = DefaultCache


    # TODO: use a named tuple, and catch when it's just the BaseParser being
    # used.
    @classmethod
    def __name_components__(klass):
        return re.findall('[A-Z][^A-Z]*', klass.__name__)

    @classmethod
    def iter_parsers(klass):
        """
        Iterate over the parsing functions defined on the parser.
        """

        for attr_name in sorted(dir(klass)):
            attr = getattr(klass, attr_name)
            if hasattr(attr, "__parser__"):
                yield attr

    @classmethod
    def iter_instances(klass, seed_kwargs, generator):
        """
        Generate instances of the parsers

        :param seed_kwargs: kwargs used to initialize the first parser
        :param generator: method to call on the parser to create
        more kwargs used to initialize additional instances. Method should
        accept a single argument with the value of `parser._read()`
        """

        queued_args = [seed_kwargs]
        used_args = []

        while len(queued_args) > 0:
            args = queued_args.pop()
            if args in used_args:
                continue
            parser = klass(**args)
            used_args.append(args)
            generate_func = getattr(klass, generator)
            new_args = generate_func(klass._read(parser))
            queued_args += new_args
            yield parser.dict()


    # TODO: use namedtuple's for the results here.
    @classmethod
    def test(klass):
        """
        Iterates over the doctests defined on a parser, returning tuples of
        (test_result, result_message). For more info see
        https://docs.python.org/2/library/doctest.html#doctestrunner-objects
        """

        doctest_runner = doctest.DocTestRunner()
        doctest_parser = doctest.DocTestParser()
        parser = klass()

        for parser_func in parser.iter_parsers():
            if parser_func.__doc__ is None:
                # We should actually alert that there's no doctest. We need our
                # own "test result" class
                continue
            string = parser_func.__doc__
            globs = {'expected': lambda: parser_func(parser.read())}
            # Shuold name refer to the name of the class here?
            name = parser.__class__.__name__
            # The filename should point to the location of the parser, not this
            # file.
            filename = __file__
            # This would be really nice. Maybe we can detect via string
            # matching?
            lineno = None
            test = doctest_parser.get_doctest(string, globs, name, filename,
                                              lineno)
            out = io.StringIO()
            test_result = doctest_runner.run(test, out=out.write)
            message = out.getvalue()
            failed = test_result.failed > 0
            yield  ParserTestResult(failed=failed, message=message,
                                    parser=parser.__class__)

    def __init__(self, lazy=False, **load_args):
        self._init_plugins()
        self.load_args = load_args
        self.all_params = self.plugin_args()
        self.all_params.update(self.load_args)
        self.cache_id = self.cache.hash_id(self.loader, **self.all_params)
        if not lazy:
            self._load()

    def __getitem__(self, key):
        for parser in self.__class__.iter_parsers():
            if parser.__name__ == key:
                return parser.__call__(self._read())
        else:
            raise KeyError

    def plugins(self):
        """
        Any class on the parser that subclasses SoupstarsPlugin is a plugin
        """

        plugins = {}
        for attr_name in dir(self.__class__):
            attr = getattr(self.__class__, attr_name)
            try:
                if issubclass(attr, SoupstarsPlugin):
                    plugins[attr_name] = attr
            except TypeError:
                pass
        else:
            return plugins


    def plugin_args(self):
        """
        Any ALL_CAPS class attribute on the parser is considered a plugin
        argument
        """

        plugin_args = {}
        for attr_name in dir(self.__class__):
            if attr_name == attr_name.upper():
                attr = getattr(self.__class__, attr_name)
                plugin_args[attr_name] = attr
        else:
            return plugin_args

    def _init_plugins(self):
        for plugin_name, plugin in self.plugins().items():
            default_plugin_args = plugin.default_args().copy()
            for arg in default_plugin_args:
                if arg in self.plugin_args():
                    default_plugin_args[arg] = self.plugin_args()[arg]
            setattr(self, plugin_name, plugin(**default_plugin_args))

    def _load(self):
        return self.cache.get(self.cache_id) \
               or self.cache.update(self.cache_id, self.loader, **self.load_args)

    def _read(self):
        return self.reader.read(self._load())

    def load(self):
        """
        Load the data to parser.
        """

        return self._load()

    def read(self):
        """
        Read the loaded data.
        """

        return self._read()

    def items(self):
        """
        Iterate over (parser_name, parse_result) values
        """

        for parser in self.__class__.iter_parsers():
            yield parser.__name__, parser.__call__(self._read())

    def dict(self):
        """
        Convert the (parser_name, parse_result) tuples to a dictionary.
        """

        return dict(self.items())

    def json(self, indent=0):
        """
        Convert the parser dictionary to json.
        """

        return json.dumps(self.dict(), indent=indent)


class HttpParser(BaseParser):
    """
    :var HOST: The default host to use, defaults to `http://0.0.0.0`
    :var ROUTE: The default route to use, defaults to `/`
    """


    # Note that default host should not end in a /
    HOST = "http://0.0.0.0"
    ROUTE = "/"

    loader = HttpLoader
    reader = HttpReader



from soupstars import HttpParser, parse, FilesystemCache


class FreeImagesParser(HttpParser):

    cache = FilesystemCache
    HOST = "https://www.pexels.com"
    ROUTE = "/photo/sunset-beach-ocean-panorama-68342/"

    @parse
    def title(soup):
        """
        >>> expected()
        'Body of Water Under Gray Clouds during Sunset'
        """
        return soup.find('h1').text

    @parse
    def url(soup):
        """
        >>> expected()
        'https://images.pexels.com/photos/68342/pexels-photo-68342.jpeg'
        """
        tag = soup.find('picture').find('img')
        src = tag['src']
        if '?' in src:
            src = src.split('?')[0]
        return src

    @parse
    def imgs(soup):
        """
        """

        imgs = soup.find_all('img')
        srcs = [t['src'] for t in imgs if '/photos/' in t['src']]
        return [s.split('?')[0] for s in srcs]

    @parse
    def id(soup):
        """
        >>> expected()
        68342
        """

        p = __class__.url(soup)
        return p.split('/')[-2]

    @parse
    def related(soup):
        results = []
        articles = soup.find_all('article')
        for art in articles:
            a = art.find('a')
            r = {'route': a.get('href')}
            results.append(r)
        return results



"""
Soupstars

~ Parse the web like a rockstar
"""

__version__ = "0.2.0"

import os

from .config import SoupstarsConfig
from .loaders import BaseLoader, DefaultLoader, HttpLoader
from .caches import BaseCache, DefaultCache, FilesystemCache, RedisCache
from .readers import BaseReader, DefaultReader, HttpReader
from .parsers import BaseParser, HttpParser
from .utils import parse, iter_parsers



from soupstars import HttpParser, parse


class NytimesArticleParser(HttpParser):

    HOST = "http://www.nytimes.com"
    ROUTE = "/2019/01/07/us/politics/trump-address-border-visit.html"

    @parse
    def title(soup):
        """
        >>> expected()
        'Trump Will Make Broad-Based Public Appeal for Border Wall'
        """

        return soup.find('h1').text.replace("\"", "")

    @parse
    def authors(soup):
        """
        >>> expected()
        'By Michael Tackett and Nicholas Fandos'
        """

        return soup.find("p", attrs={"itemprop": "author creator"}).text.replace("\"", "")

    @parse
    def published_at(soup):
        """
        >>> expected()
        'Jan. 7, 2019'
        """

        return soup.find("time").text.replace("\"", "")



"""
Configuration
-------------

Configuration objects work like a dictionary, with values determined by
environment variables.

>>> import os
>>> from soupstars.config import SoupstarsConfig
>>> config = SoupstarsConfig()
"""

import os
import importlib


class SoupstarsConfig(object):
    """
    Determines the configuration values:

    :var SOUPSTARS_HOME: The directory used to store parsers and caches.
        Defaults to `~/.soupstars`.
    :var PARSERS_PATH: The directory searched for additional parsers. Defaults
        to `~/.soupstars/parsers`.
    :var LOAD_EXAMPLES: Whether to include the example parsers installed with
        soupstars when running the flask server.
    :var FILESYSTEM_CACHE_PATH: The directory used for storing cached objects
        by `FilesystemCache`. Defaults to `~/.soupstars/cache`.
    """

    def __init__(self):
        self.SOUPSTARS_HOME = self._set_soupstars_home()
        self.SOUPSTARS_CACHE_PATH = self._set_cache_path()
        self.SOUPSTARS_LOAD_EXAMPLES = self._set_load_examples()
        self.SOUPSTARS_EXAMPLES_PATH = self._set_soupstars_example_path()
        self._set_directories()

    def _set_soupstars_home(self):
        home = os.environ.get('SOUPSTARS_HOME') \
               or os.path.join(os.path.expanduser('~'), 'soupstars')
        return home

    def _set_load_examples(self):
        if os.environ.get("SOUPSTARS_LOAD_EXAMPLES") in ('False', 0):
            return False
        else:
            return True

    def _set_soupstars_example_path(self):
        _soupstars_path = os.path.abspath(os.path.dirname(__file__))
        SOUPSTARS_EXAMPLES_PATH = os.path.join(_soupstars_path, "examples")
        return SOUPSTARS_EXAMPLES_PATH

    def _set_cache_path(self):
        return os.path.join(self.SOUPSTARS_HOME, 'cache')

    def _set_directories(self):
        for path in [self.SOUPSTARS_HOME, self.SOUPSTARS_CACHE_PATH]:

            try:
                os.mkdir(path)
            except FileExistsError:
                pass



"""
Loaders
-------

The loader classes are intended to interface with other systems, and are
created by subclassing the `BaseLoader` class. A loader should expose a
`load` method, which is generally used by the `Cache` objects whenever
it can't find an object it has already stored.

In the simplest case, the `DefaultLoader` just returns the arguments it
receives.

>>> from soupstars import DefaultLoader
>>> loader = DefaultLoader()
>>> loader.load(a=1, b=2)
{'a': 1, 'b': 2}

The `HttpLoader` will pull data from a website. The class attributes
`default_route` and `default_host` are used to control how the loader builds
its requests.

>>> loader = HttpLoader()
>>> loader.ROUTE
'/photos/1'
>>> from minimock import Mock
>>> requests.get = Mock("requests.get")
>>> requests.get.mock_returns = "result"
>>> loader.load()
Called requests.get('https://jsonplaceholder.typicode.com/photos/1')
'result'
>>> requests.get.mock_returns = "result2"
>>> loader.load(url='https://www.google.com/search?')
Called requests.get('https://www.google.com/search?')
'result2'
>>> requests.get.mock_returns = "result3"
>>> loader.load(route="/photos/2")
Called requests.get('https://jsonplaceholder.typicode.com/photos/2')
'result3'

"""

import requests

from .exceptions import NotImplementedError
from .utils import SoupstarsPlugin


class BaseLoader(SoupstarsPlugin):
    """
    Base loader. Subclass this to create your own loader.
    """

    def load(self, **kwargs):
        raise NotImplementedError


class DefaultLoader(BaseLoader):
    """
    Default loader, returns the keyword arguments passed to `load`.
    """

    def load(self, **kwargs):
        return kwargs


class HttpLoader(BaseLoader):
    """
    Uses requests.get to load a webpage.
    """

    HOST = "https://jsonplaceholder.typicode.com"
    ROUTE = "/photos/1"

    # Move the doctests into the module docstring
    def load(self, **kwargs):
        if 'url' not in kwargs and 'route' not in kwargs:
            return requests.get(self.HOST + self.ROUTE, **kwargs)
        elif 'url' in kwargs:
            url = kwargs.pop('url')
            return requests.get(url, **kwargs)
        elif 'route' in kwargs:
            route = kwargs.pop('route')
            return requests.get(self.HOST + route, **kwargs)
        else:
            raise AttributeError("Unexpected argument handling")



"""
TODO:
    - ls command (parsers under home, examples, etc)
    - open command (for finding parsers)
    - create command
    - parse command
        - pass additional command line args to parser's init (ie to loader)
    - test command
        - prettier formatting of test results
        - named tuples for test results
    - search command
        - do some scanning for specific words in the `read` object
    - show command
        - describe the parsing functions on the parser
"""

import datetime as dt
import inspect
import os
import re

import click
from click import echo, style

from .utils import iter_parsers


LOGGING_STRING = "{ts} [{level}] {text}"


def info(text):
    level = style("INF", fg="blue", bold=True)
    ts = style(str(dt.datetime.now().time()), fg="bright_white", bold=True)
    echo(LOGGING_STRING.format(level=level, text=text, ts=ts))


def success(text):
    level = style("SUC", fg="green", bold=True)
    ts = style(str(dt.datetime.now().time()), fg="bright_white", bold=True)
    echo(LOGGING_STRING.format(level=level, text=text, ts=ts))


def fail(text):
    level = style("ERR", fg="red", bold=True)
    ts = style(str(dt.datetime.now().time()), fg="bright_white", bold=True)
    echo(LOGGING_STRING.format(level=level, text=text, ts=ts))


@click.group()
def main():
    pass


@main.command()
def ls():
    """
    List all available parsers
    """

    for Parser in iter_parsers():
        name = Parser.__name__
        parser_type = style(Parser.__base__.__name__.split('.')[-1], fg="cyan", bold=True)
        location = Parser.__module__
        text = "\t".join([name, parser_type, location])
        info(text)

# The arguments here should be optional filters to apply to the name of
# the parser. By default `soupstars test` should run tests on all parsers.
@main.command()
@click.argument('scope')
@click.argument('name')
def test(scope, name):
    """
    Test the `@parse` methods from a parser using its docstrings
    """

    for Parser in iter_parsers():
        name_components = [s.lower() for s in re.findall('[A-Z][^A-Z]*', Parser.__name__)]
        pscope, pname, _parser_str = name_components
        if pscope == scope.lower() and pname == name.lower():
            for test_result in Parser().test():
                if test_result.failed:
                    fail("Failed")
                    fail(test_result.message)
                else:
                    success("Passed")


@main.command()
@click.argument('scope')
@click.argument('name')
def parse(scope, name):
    """
    Run a parser
    """
    for Parser in iter_parsers():
        name_components = [s.lower() for s in re.findall('[A-Z][^A-Z]*', Parser.__name__)]
        pscope, pname, _parser_str = name_components
        if pscope == scope.lower() and pname == name.lower():
            info(f"Running {Parser.__name__}")
            success("Received: \n" + Parser().json(indent=2))


@main.command()
def dirs():
    """
    List directories with parsers
    """

    # print('tset')
    from soupstars.utils import iter_parser_directories
    for dir_ in iter_parser_directories():
        info(str(dir_))


@main.command()
def packages():
    """
    List packages with parsers
    """

    # print('tset')
    from soupstars.utils import iter_package_paths
    for pp in iter_package_paths():
        info(str(pp))


@main.command()
def modules():
    """
    List the modules searched for parsers
    """

    from soupstars.utils import iter_parser_modules
    for pp in iter_parser_modules():
        info(str(pp))


@main.command()
@click.argument('package')
@click.argument('module')
def create(package, module):
    """
    Create a new parser in the soupstars home directory
    """

    from soupstars import SoupstarsConfig
    config = SoupstarsConfig()

    parser_template = """from soupstars import HttpParser, parse, FilesystemCache

class {package}{module}Parser(HttpParser):

    cache = FilesystemCache
    HOST = "https://www.facebook.com"
    ROUTE = "/"

    @parse
    def title(soup):
        \"\"\"
        >>> expected()
        "My title"
        \"\"\"
        return soup.find('h1')

    """.format(package=package.capitalize(), module=module.capitalize())

    package_path = os.path.join(config.SOUPSTARS_HOME, package)
    module_path = os.path.join(package_path, module) + '.py'
    try:
        os.mkdir(package_path)
    except FileExistsError:
        pass


    f = open(os.path.join(package_path, '__init__.py'), 'w+')
    f.write('"test"')
    f.close()

    try:
        with open(module_path, 'w') as f:
            f.write(parser_template)
    except Exception:
        fail("Failed")



"""
Caches
------

Caches help reduce unnecessary loading requests. The `DefaultCache` is just a
a python dictionary with some additional features.

>>> from soupstars import DefaultCache
>>> cache = DefaultCache()
>>> cache.set('key', 'value')
>>> cache.get('key')
'value'

Unlike dictionary, caches provide a hashing function that transforms a `Loader`
along with its `load` keyword arguments into a unique string. This
allows the cache to uniquely identify a `Loader`'s response without
actually calling the `load` method of the `Loader`.

>>> from soupstars import DefaultLoader
>>> loader = DefaultLoader()
>>> id = cache.hash_id(loader, a=1, b=2)
>>> id
'ac6dcde3d009f3b5d2ea803c12ef1a3c'

Caches also allow updating a key using a `Loader` instance directly.

>>> cache.update('key', loader, a=1, b=2)
{'a': 1, 'b': 2}
>>> cache.get('key')
{'a': 1, 'b': 2}

For a more persistent cache, you can use the `FilesystemCache` to store
responses forever. If you need a cache available to multiple nodes, you can
use the RedisCache, which stores `load` responses for a single day.
"""

import hashlib
import redis
import pickle
import os

from .exceptions import NotImplementedError
from .config import SoupstarsConfig
from .utils import SoupstarsPlugin


config = SoupstarsConfig()

class BaseCache(SoupstarsPlugin):
    """
    Basic cache definition. Any subclass must define get and set methods.
    """

    def get(self, key):
        raise NotImplementedError

    def set(self, key, value):
        raise NotImplementedError

    # I think there might be a better hashing strategy. See this:
    # https://stackoverflow.com/questions/5884066/hashing-a-dictionary
    def hash_id(self, loader, **kwargs):
        """
        Uses the class and keyword args to generate a unique id for a
        loader. Can receive any other keyword args, such as a loader's init
        args, to make the id unique
        """

        load_args = [':'.join(str(item) for item in kwargs.items())]
        if load_args:
            load_args = '_'.join(load_args)
        else:
            load_args = ""

        name_value = str(loader.__class__)
        string_value = " | ".join([load_args, name_value])
        hash_value = hashlib.md5(string_value.encode('utf8'))
        return hash_value.hexdigest()

    def update(self, key, loader, **loader_args):
        """
        Updates the cache's `key` value using the response from calling
        `loader.load(**loader_args)`.
        """

        load = loader.load(**loader_args)
        self.set(key, load)
        return load


class DefaultCache(BaseCache):
    """
    A simple in-memory cache. Expires at the end of a python session.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._cache = {}

    def get(self, key):
        return self._cache.get(key)

    def set(self, key, value):
        self._cache[key] = value


class RedisCache(BaseCache):
    """
    A redis backed cache. Stores items for one day. Using the cache requires
    the environment variable `SOUPSTARS_REDIS_URL` set. Cached items are stored
    as python `Pickle` objects.
    """

    # TODO: use config here
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        try:
            self.URL = config['SOUPSTARS_REDIS_URL']
        except KeyError:
            msg = """Couldn't find a redis url in the environment. You need to
            set SOUPSTARS_REDIS_URL in order to use a redis backend.
            """
            raise KeyError(msg)

        self._cache = redis.Redis.from_url(self.URL)

    def get(self, key):
        pickled_value = self._cache.get(key)
        if pickled_value is None:
            return None
        return pickle.loads(pickled_value)

    def set(self, key, value):
        pickled_value = pickle.dumps(value)
        self._cache.set(key, pickled_value, ex=60*60*24)  # 1 day


class FilesystemCache(BaseCache):
    """
    A filesystem backed cache. Stores items forever.

    Using this cache will create a folder under $HOME/.soupstars. If the
    process does not have write access to $HOME, the cache will fail.
    """

    # TODO: use config here
    def __init__(self, *args, **kwargs):
        home_dir = os.path.expanduser('~')
        cache_dir = os.path.join(home_dir, 'soupstars', 'cache')
        if not os.path.exists(cache_dir):
            os.mkdir(cache_dir)
        self.URL = cache_dir

    def get(self, key):
        filename = os.path.join(self.URL, key + '.cache')
        if not os.path.exists(filename):
            return None
        else:
            with open(filename, 'rb') as file_handle:
                return pickle.load(file_handle)

    def set(self, key, value):
        filename = os.path.join(self.URL, key + '.cache')
        with open(filename, 'wb') as file_handle:
            pickle.dump(value, file_handle)



from soupstars import HttpParser, parse, FilesystemCache


class BibleVersesParser(HttpParser):

    cache = FilesystemCache
    HOST = "http://topverses.com/Bible"
    ROUTE = "/&pg=1&a=ajax"

    @parse
    def sections(soup):
        return [t.text.strip() for t in soup.find_all('h2')]

    @parse
    def ranks(soup):
        rank_texts = [t.text.strip() for t in soup.find_all('h5')]
        rank_nums = [t.split()[-1] for t in rank_texts]
        return [int(t) for t in rank_nums]

    @parse
    def quotes(soup):
        containers = soup.find_all('div', attrs={'class': 'container'})
        sections = [c.find('div', attrs={'class': 'slide section'}) for c in containers]
        texts = [s.find('div').text.replace('NIV', '') for s in sections]
        clean_texts = [s.strip() for s in texts]
        return [str(c) for c in clean_texts]

    @parse
    def related(soup):
        return [{"route": "/&pg={n}&a=ajax".format(n=n)} for n in range(1,11)]
